//
// Created by johann on 13/08/21.
// The idea of this class is to generate a wrapper class around Csound and Oboe to create audio output callbacks when starting Csound.
// This class will be called by the implementation of CsoundDart.
//

#include<iostream>
#include"include/csound.h"
#include<oboe/Oboe.h>
#include<memory>
#include <android/log.h>
#include<thread>
#include<sstream>
#include"CallbackManager.h"
#include<atomic>

extern "C"
{
#define APPNAME "csounddart_example"

__attribute__((visibility("default"))) __attribute__((used))
void printC(const char *to_print)
{
    __android_log_print(ANDROID_LOG_DEBUG, APPNAME, "CsoundAndroid CPP -> %s", to_print);
}

/*
* CsoundAndroid - Csound C++ API with Oboe for Android
*/

class CsoundAndroid : public oboe::AudioStreamCallback
{
public:
    CsoundAndroid() : isPlaying(false)
    {
        _cs = csoundCreate(NULL);
    }

    ~CsoundAndroid()
    {
        csoundDestroy(_cs);
    }

    int CompileCsdText(const char *csd_text)
    {
        return csoundCompileCsdText(_cs, csd_text);
    }

    int CompileCsd(const char *csd_path)
    {
        return csoundCompileCsd(_cs, csd_path);
    }

    int CompileOrc(const char *orchestra)
    {
        return csoundCompileOrc(_cs, orchestra);
    }

    int ReadScore(const char *score)
    {
        return csoundReadScore(_cs, score);
    }

     oboe::DataCallbackResult onAudioReady(oboe::AudioStream *oboeStream,
                void *audio_data,
                int32_t frame_count)
     {
        isPlaying = true;
        int res = 0;
        res = PerformKsmps();
        int ksmps = GetKsmps();
        int nchnls = GetNchnls();
        MYFLT *spout = csoundGetSpout(_cs);
        float *float_buffer = static_cast<float *>(audio_data);
        for (int i = 0; i < ksmps; i++) {
            for (int j = 0; j < nchnls; j++) {
                float_buffer[i * nchnls + j] = spout[i * nchnls + j];
            }
         }
        if(send_port != -1) {
            callbackToDartInt32(send_port, res);
        }
        return oboe::DataCallbackResult::Continue;
     }

    int Start()
    {
        printC("Starting");
        int res =  csoundStart(_cs);

        oboe::AudioStreamBuilder builder;

        builder.setCallback(this);
        builder.setSharingMode(oboe::SharingMode::Exclusive);
        builder.setPerformanceMode(oboe::PerformanceMode::LowLatency);
        builder.setSampleRate(GetSr());
        builder.setFramesPerCallback(GetKsmps());
        builder.setChannelCount(GetNchnls());
        builder.setFormat(oboe::AudioFormat::Float);

        oboe::Result result = builder.openStream(audio_stream);
        std::thread t([=](){
            audio_stream->requestStart();
        });
        t.detach();

        return res;
    }

    int PerformKsmps()
    {
        return csoundPerformKsmps(_cs);
    }


    int GetNchnls()
    {
        return csoundGetNchnls(_cs);
    }

    int GetNchnlsInput()
    {
        return csoundGetNchnlsInput(_cs);
    }

    int GetKsmps()
    {
        return csoundGetKsmps(_cs);
    }

    double GetSr()
    {
        return csoundGetSr(_cs);
    }

    double Get0dBFS()
    {
        return csoundGet0dBFS(_cs);
    }

    void Stop()
    {
        if(audio_stream)
        {
            audio_stream->requestStop();
        }
        csoundStop(_cs);
    }

    void SetControlChannel(const char *channelName, double value)
    {
        csoundSetControlChannel(_cs, channelName, value);
    }

    double GetControlChannel(const char *channelName)
    {
        return csoundGetControlChannel(_cs, channelName, NULL);
    }

    void GetAudioChannel(const char *name, float *audio)
    {
        //MYFLT *float_buffer = static_cast<MYFLT *>(audio);
        csoundGetAudioChannel(_cs, name, audio);
    }

    int SetOption(const char *opt)
    {
        return csoundSetOption(_cs, opt);
    }

    void Pause(bool pause)
    {
        if(audio_stream) {
            if(pause)
            {
                audio_stream->requestPause();
            } else
            {
                audio_stream->requestStart();
            }
        }
    }


    Dart_Port send_port = -1;
private:

    std::atomic<bool> isPlaying;
    CSOUND *_cs;
    oboe::ManagedStream managedStream;
    std::shared_ptr<oboe::AudioStream> audio_stream;
};



// C implementation for the C++ class CsoundAndroid
__attribute__((visibility("default"))) __attribute__((used))
CsoundAndroid *csoundAndroidCreate()
{
    printC("csound Create");
    return new CsoundAndroid();
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidCompileCsdText(CsoundAndroid *cs, const char *csd_text)
{
    printC("compile Csd Text");
    return cs->CompileCsdText(csd_text);
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidStart(CsoundAndroid *cs)
{
    printC("csound Start");
    return cs->Start();
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidPerformKsmps(CsoundAndroid *cs)
{
    return cs->PerformKsmps();
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidDestroy(CsoundAndroid *cs)
{
    delete cs;
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidStop(CsoundAndroid *cs)
{
    cs->Stop();
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidSetControlChannel(CsoundAndroid *cs, const char *channelName, double value)
{
    cs->SetControlChannel(channelName, value);
}
__attribute__((visibility("default"))) __attribute__((used))
double csoundAndroidGetControlChannel(CsoundAndroid *cs, const char *channelName)
{
    return cs->GetControlChannel(channelName);
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidGetAudioChannel(CsoundAndroid *cs, const char *name, float *data)
{
    cs->GetAudioChannel(name, data);
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidCompileCsd(CsoundAndroid *cs, const char *csd)
{
    return cs->CompileCsd(csd);
}
__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidCompileOrc(CsoundAndroid *cs, const char *orc)
{
    return cs->CompileOrc(orc);
}
__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidReadScore(CsoundAndroid *cs, const char *score)
{
    return cs->ReadScore(score);
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidSetCallbackPort(CsoundAndroid * cs, int64_t dart_port)
{
    std::string s = "setting dart port Cside " + std::to_string(dart_port);
    printC(s.c_str());
    cs->send_port = dart_port;
    printC("dart port set ok ");
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidGetKsmps(CsoundAndroid *cs)
{
    return cs->GetKsmps();
}
__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidGetNchnls(CsoundAndroid *cs)
{
    return cs->GetNchnls();
}
__attribute__((visibility("default"))) __attribute__((used))
double csoundAndroidGetSr(CsoundAndroid *cs)
{
    return cs->GetSr();
}
__attribute__((visibility("default"))) __attribute__((used))
double csoundAndroidGet0dBFS(CsoundAndroid *cs)
{
    return cs->Get0dBFS();
}

__attribute__((visibility("default"))) __attribute__((used))
int csoundAndroidSetOption(CsoundAndroid *cs, const char *opt)
{
    return cs->SetOption(opt);
}

__attribute__((visibility("default"))) __attribute__((used))
void csoundAndroidPause(CsoundAndroid *cs, int pause)
{
    cs->Pause(pause == 1);
}

}
