@JS()
library csound;

import 'csound.dart';
import 'package:js/js.dart';
import 'package:js/js_util.dart' show promiseToFuture;

Future<void> importLibs() async
{
}

/* Removed in Csound 6.15
@JS("importScripts")
external dynamic importScripts(String s);


Future importJsScripts(String s) {
  return promiseToFuture(importScripts(s));
}
*/

/*
JS Wrappers around CsoundObj and CsoundDart.js (a glue code file linked in the plugin).
 */
@JS("resumeAudioState")
external void resumeAudioState();
@JS("suspendAudioState")
external void suspendAudioState();

@JS("csoundInitialize")
external dynamic csoundInitialize();

Future initializeCsound() {
  print("initializeCsound");
  return promiseToFuture(csoundInitialize());
}

@JS("CsoundJsDart")
class CsoundJsDart
{
  // Constructor
 external CsoundJsDart();

 external initialize();
  // start and play both start engine and play performance
 external start();
 external play();
 external stop();
 external reset();

  //set option string to Csound
 external setOption(String optionString);

 external compileCSD(String csdString);
 external compileOrc(String orcString);
 external readScore(String score);

 // not fully working now
 external evaluateCode(String codeString);


 external setControlChannel(String channelName, double value);
 external setStringChannel(String channelName, String stringValue);


 external num getControlChannel(String channelName);
 external String getStringChannel(String channelName);

  // callback to perform when channel is available
 external requestControlChannel(String channelName, Function callback);
 external requestStringChannel(String channelName, Function callback);

 external setMessageCallback(Function msgCallback);

  // receives a callback function taking a number as argument (channel value)
  // need to use allowInterop to pass callback function
 external setControlChannelCallbackJs(String channelName, Function callback);
 external setStringChannelCallbackJs(String channelName, Function callback);


 external setTable(int table, List<double> value);
 external setTableCallbackJs(int table, Function callback);
}

/*
The implementation of Csound for web.
 */
class CsoundWeb implements Csound
{
  CsoundWeb() {
    print("init Csound");
    initializeCsound().then((value) {
      resumeAudioState();
      _csound = CsoundJsDart();
    });
  }


  Future<int> compileCsdText(String text)
  {
    _csound.compileCSD(text);
    return Future<int>.value(0);
  }

  Future<int> compileOrc(String orc)
  {
    _csound.compileOrc(orc);
    return Future<int>.value(0);
  }

  Future<int> compileCsd(String csd)
  {
    _csound.compileCSD(csd);
    return Future<int>.value(0);
  }

  Future<int> readScore(String score)
  {
    _csound.readScore(score);
    return Future<int>.value(0);
  }

  int perform()
  {
    _csound.start();
    return 0;
  }

  void pause(bool b)
  {
    throw(UnimplementedError());
  }

// Web implementation of Csound (CsoundObj) seems threaded by default, to verify

  void stop()
  {
    _csound.stop();
  }

  void setControlChannel(String channel, double val)
  {
    _csound.setControlChannel(channel, val);
  }

  Future<double> getControlChannel(String channel) async
  {
    return _csound.getControlChannel(channel).toDouble();
  }

  void setControlChannelCallback(String channelName, Function callback)
  {
    _csound.setControlChannelCallbackJs(channelName, callback);
  }

  void setStringChannel(String channelName, String value)
  {
    _csound.setStringChannel(channelName, value);
  }
  
  Future<String> getStringChannel(String channelName) async 
  {
    return _csound.getStringChannel(channelName);
  }

  void setStringChannelCallback(String channelName, Function callback) 
  {
    _csound.setStringChannelCallbackJs(channelName, callback);
  }


  void setTable(int table, List<double> value)
  {
    _csound.setTable(table, value); 
  }

  void setTableCallback(int table, Function callback)
  {
    _csound.setTableCallbackJs(table, callback);
  }

  void setAudioChannelCallback(String channel, Function f)
  {
    throw UnimplementedError("Not implemented for web");
  }
  //member


  late CsoundJsDart _csound;

  @override
  Future<int> getKsmps() {
    // TODO: implement getKsmps
    throw UnimplementedError("Not impleted for web");
  }

  @override
  Future<int> getNchnls() {
    // TODO: implement getNchnls
    throw UnimplementedError("Not impleted for web");
  }

  @override
  Future<int> getSr() {
    // TODO: implement getSr
    throw UnimplementedError("Not impleted for web");
  }

  @override
  Future<double> evalCode(String code) {
    // TODO: implement evalCode
    throw UnimplementedError();
  }

  @override
  Future<int> get0dbfs() {
    // TODO: implement get0dbfs
    throw UnimplementedError();
  }

  Future<int> setOption(String opt)
  {
    throw UnimplementedError();
  }
}

Csound getCsoundDart() {
  importLibs();
  return CsoundWeb();
}

Function getAppStartup()
{
  return () {};
}